"""
PyPDC
=========

PDC is a Python package for calculating asymptotic Partial Directed Coherence estimations for brain connectivity analysis.

"""

name = "pypdc"

#Current version:
__version__ = "0.0.6"




